# coding=utf-8
#
# Standard libraries
import json
import pkgutil

# Third party libraries
from textsearch import TextSearch

# etltk libraries


def _load_json_data(name: str):
    """
    Load a json content from ./lang/am/data/``name``.json and return it.
    """
    json_data = pkgutil.get_data("etltk.lang", "am/data/{0}.json".format(name))
    return json.loads(json_data.decode("utf-8"))

def _replace(text: str, ts_replacer: TextSearch) -> str:
    return ts_replacer.replace(text)

# Load the Amharic character levels replacers dictionary
char_replacers_dict = _load_json_data("char_replacers_dict")

# Load the Amharic punctuation replacers dictionary
punct_replacers_dict = _load_json_data("punct_replacers_dict")

# Load the Amharic character labialized dictionary
labialized_dict = _load_json_data("labialized_dict")

# Load the Amharic short form expansions dictionary
shortened_expansions_dict = _load_json_data("shortened_expansions_dict")

# Update the `shortened_expansions_dict` to support both `.` and `/`` patterns like አ.አ and አ/አ
shortened_expansions_dict.update({k.replace(".", "/"): v for k, v in shortened_expansions_dict.items()})

ts_char_replacer = TextSearch("insensitive", "norm")
ts_char_replacer.add(char_replacers_dict)

ts_labialized_replacer = TextSearch("insensitive", "norm")
ts_labialized_replacer.add(labialized_dict)

ts_expand_shortened_replacer = TextSearch("insensitive", "norm")
ts_expand_shortened_replacer.add(shortened_expansions_dict)

ts_punct_replacer = TextSearch("insensitive", "norm")
ts_punct_replacer.add(punct_replacers_dict)

def normalize_char(text: str) -> str:
    # Character Level Normalization 
    # such as ጸሀይ and ፀሐይ.
    return _replace(text, ts_replacer=ts_char_replacer)

def normalize_punct(text: str) -> str:
    # Punctuation Normalization 
    # such as :: to ።.
    return _replace(text, ts_replacer=ts_punct_replacer)

def normalize_labialized(text: str) -> str:
    # Labialized Character Normalzation 
    # such as ሞልቱዋል to ሞልቷል
    return _replace(text, ts_replacer=ts_labialized_replacer)

def normalize_shortened(text: str) -> str:
    # Short Form Expansion 
    # such as ጠ/ሚ to ጠቅላይ ሚኒስተር.
    return _replace(text, ts_replacer=ts_expand_shortened_replacer)


# coding: utf8
from __future__ import unicode_literals

# Standard libraries
import re
import string

# Third party libraries
import emoji

# etltk libraries
from .punctuation import ASSCII_PUNCT, ETHIOPIC_PUNCT, NO_ABBREV_ASSCII_ETHIOPIC_PUNCTS
from .utils import is_chinese_char, is_ethiopic, is_ethiopic_digit, regex_replace
     
# Regular expression
REGEX_PATTERN_URLS = re.compile(r'https?://\S+|www\.\S+')
REGEX_PATTERN_TAGS = re.compile(r"(?x)<[^>]+>| &([a-z0-9]+|\#[0-9]{1,6}|\#x[0-9a-f]{1,6});")
REGEX_PATTERN_EMAIL = re.compile(r'[a-z0-9._%+-]+@[a-z0-9.-]+\.[a-z]{2,}')
REGEX_ETHIOPIC_PUNCTUATION = re.compile(ETHIOPIC_PUNCT)
REGEX_PATTERN_ASCII = re.compile(r'[A-Za-z]+')
REGEX_PATTERN_DIGITS = re.compile(r"\d+")
REGEX_PATTERN_ARABIC = re.compile('([\u0621-\u064A]+)')
# TODO: add more special characters
SPECIAL_CHARACTERS = 'å¼«¥ª°©ð±§µæ¹¢³¿®ä£"”“`‘´’‚,„»«「」『』（）〔〕【】《》〈〉'

def remove_whitespaces(text: str) -> str:
    """Remove extra spaces, tabs, and new lines 
    from a text string
    """
    return " ".join(text.split())

def remove_links(text: str) -> str:
    """Remove URLs from a text string
    """
    return regex_replace(text, pattern=REGEX_PATTERN_URLS, replace='')

def remove_tags(text: str) -> str:
    """Remove HTML tags from a text string
    """
    return regex_replace(text, pattern=REGEX_PATTERN_TAGS, replace='')

def remove_emojis(text: str) -> str:
    """Remove emojis from a text string
    """
    return emoji.replace_emoji(text, replace='')

def remove_email(text: str) -> str:
    """Remove email adresses from a text string
    """
    return regex_replace(text, pattern=REGEX_PATTERN_EMAIL, replace='')

def remove_punct(text: str, abbrev: bool = True):
    # TODO: remove punctuations like ዓ.ም.
    if not abbrev:
        # List of punctuation includes ethiopic short form punctuations `.` and `/`
        string_punctuations = ASSCII_PUNCT + ETHIOPIC_PUNCT
    else:
        # List of punctuation excluded ethiopic short form punctuations `.` and `/`
        # remove `.` and `/` punctuation from punctuations
        string_punctuations = NO_ABBREV_ASSCII_ETHIOPIC_PUNCTS

    return _remove_punct(text, punctuation=string_punctuations)

def _remove_punct(text: str, punctuation) -> str:
    """Remove punctuations from a text string
    """
    if not len(punctuation.strip()):
        raise ValueError(f"{__name__}: `punctuation` can't be `empty string`")
    
    # Split into words by white space
    words = text.split()

    # Remove punctuation from each word
    table = str.maketrans('', '', punctuation)

    return " ".join([w.translate(table) for w in words])

def remove_ethiopic_punct(text: str) -> str:
    """Remove ethiopic punctuations from a text string
    """
    return regex_replace(text, pattern=REGEX_ETHIOPIC_PUNCTUATION, replace='')

def remove_special_characters(text: str) -> str:
    """Removes special characters from a text string
    """
    return text.translate(str.maketrans('', '', SPECIAL_CHARACTERS))

def remove_digits(text: str):
    """Remove all digits from a text string
    """
    return regex_replace(text, pattern=REGEX_PATTERN_DIGITS, replace='')

def remove_ethiopic_digits(text):
    """Remove all ethiopic digits from a text string
    """
    output = [
        char for char in text if not is_ethiopic_digit(char)
    ]
    return "".join(output)

def remove_english_chars(text: str) -> str:
    """Remove ascii characters from a text string
    """
    return regex_replace(text, pattern=REGEX_PATTERN_ASCII, replace='')

def remove_arabic_chars(text: str) -> str:
    """Remove arabic characters and numerals from a text string
    """
    return regex_replace(text, pattern=REGEX_PATTERN_ARABIC, replace='')

def remove_chinese_chars(text: str) -> str:
    """Remove chinese characters from a text string
    """
    output = [
        char for char in text if not is_chinese_char(ord(char))
    ]
    return "".join(output)

def remove_non_ethiopic(text: str):
    """Remove non ethioipc characters from a text string
    """
    ethiopic_only = "".join([char for char in text if (is_ethiopic(char) or ord(char) == 32)])
    return ethiopic_only


# coding=utf-8
#
# Standard libraries
import re
import unicodedata
from typing import List

# etltk libraries

# Line Tokenizer
class LineTokenizer(object):
    def tokenize(self, text: str):
        """Tokenize text by newline.
        """
        lines = [
            line for line in text.splitlines() if line != ""
        ]
        return lines



# coding=utf-8
#
# Standard libraries
import re
import unicodedata

def is_ethiopic(char):
    cp = ord(char)
    if (cp >= 0x1200 and cp <= 0x137F):
        return True
    
    return False

def is_ethiopic_digit(char):
    cp = ord(char)
    if ((cp >= 0x1369 and cp <= 0x136F) or  # ፩ - ፯ 
        (cp >= 0x1370 and cp <= 0x137C)):   # ፰ - ፼
        return True
    
    return False

def is_ethiopic_punct(char) -> bool:
    """Checks whether `chars` is a amharic punctuation character."""
    cp = ord(char)
    # Characters such as ፠ ፡ ። ፣ ፤ ፥ ፦ ፧ ፨ 
    if (cp >= 0x1360 and cp <= 0x1368):
        return True
    return False

def is_special_punct(char):
    """Checks whether `chars` is a special punctuation character."""
    cp = ord(char)

    # Characters such as . /
    # In ethiopic these chars are use to create short froms 
    # like `ዓ.ም` `ዓ/ም`        
    if (cp >= 46 and cp <= 47):
        return True
    return False

def is_english_punct(char) -> bool:
    """Checks whether `chars` is a punctuation character."""
    cp = ord(char)
    # We treat all non-letter/number ASCII as punctuation.
    # Characters such as "^", "$", and "`" are not in the Unicode
    # Punctuation class but we treat them as punctuation anyways, for
    # consistency.
    if ((cp >= 33 and cp <= 47) or (cp >= 58 and cp <= 64) or
            (cp >= 91 and cp <= 96) or (cp >= 123 and cp <= 126)):
        return True
    cat = unicodedata.category(char)
    if cat.startswith("P"):
        return True
    return False

def is_control(char) -> bool:
    """Checks whether `chars` is a control character."""
    if char == "\t" or char == "\n" or char == "\r":
        return True
    return False

def remove_control_char(s):
    """Remove a control character in a string."""
    return "".join(ch for ch in s if unicodedata.category(ch)[0]!="C")

def is_chinese_char(cp) -> bool:
    """Checks whether CP is the codepoint of a CJK character."""
    # This defines a "chinese character" as anything in the CJK Unicode block:
    #   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
    #
    # Note that the CJK Unicode block is NOT all Japanese and Korean characters,
    # despite its name. The modern Korean Hangul alphabet is a different block,
    # as is Japanese Hiragana and Katakana. Those alphabets are used to write
    # space-separated words, so they are not treated specially and handled
    # like the all of the other languages.
    if ((cp >= 0x4E00 and cp <= 0x9FFF) or  #
            (cp >= 0x3400 and cp <= 0x4DBF) or  #
            (cp >= 0x20000 and cp <= 0x2A6DF) or  #
            (cp >= 0x2A700 and cp <= 0x2B73F) or  #
            (cp >= 0x2B740 and cp <= 0x2B81F) or  #
            (cp >= 0x2B820 and cp <= 0x2CEAF) or
            (cp >= 0xF900 and cp <= 0xFAFF) or  #
            (cp >= 0x2F800 and cp <= 0x2FA1F)):  #
        return True

    return False


def regex_replace(text: str, pattern: str, replace: str = '') -> str:
    """ Uses a regular expression to perform substitution on a sequence of characters. """
    return pattern.sub(replace, text, re.UNICODE)


'''
from . import amharic
amharic_preprocessor = amharic.preprocessing
amharic_normalizer = amharic.normalization
amharic_tokenizer = amharic.tokenization

from .amharic import (    
    clean_amharic,
    normalize_amharic
)

from ethltk.oromiffa import (
    OromiffaPreprocessor,
    clean_oromiffa
)
'''

import os
from .lang.am import AmharicDocument
from .common.doc import Sentence, Word, WordList

__version__ = '0.0.10'
__license__ = 'MIT'
__author__ = 'Robel Equbasilassie'

PACKAGE_DIR = os.path.dirname(os.path.abspath(__file__))

__all__ = [
    'AmharicDocument',
    'Word',
    'WordList',
    'Sentence'
]



class Word(str):
    """A simple Amharic word representation.
    """
    def __new__(cls, string):
        """Return a new instance of the class.
        """
        return super(Word, cls).__new__(cls, string)
    
    def __init__(self, string):
        self.string = string

    def __repr__(self):
        return repr(self.string)

    def __str__(self):
        return self.string
    
class WordList(list):
    """A list-like collection of words."""
     
    def __init__(self, collection):
        """Initialize a WordList. Takes a collection of strings as
        its only argument.
        """
        super(WordList, self).__init__([Word(w) for w in collection])
    
    def __str__(self):
        """Returns a string representation for printing."""
        return super(WordList, self).__repr__()
    
    def __repr__(self):
        """Returns a string representation for debugging."""
        class_name = self.__class__.__name__
        return '{cls}({lst})'.format(cls=class_name, lst=super(WordList, self).__repr__())

    def __getitem__(self, key):
        """Returns a string at the given index."""
        item = super(WordList, self).__getitem__(key)
        if isinstance(key, slice):
            return self.__class__(item)
        else:
            return item
    
    def __setitem__(self, index, obj):
        """Places object at given index, replacing existing item. If the object
        is a string, inserts a :class:`Word <Word>` object.
        """
        if isinstance(obj, (str, bytes)):
            super(WordList, self).__setitem__(index, Word(obj))
        else:
            super(WordList, self).__setitem__(index, obj)

    def count(self, strg, *args, **kwargs):
        """Get the count of a word or phrase `s` within this WordList.

        :param strg: The string to count.
        """
        return super(WordList, self).count(strg, *args, **kwargs)

    def append(self, obj):
        """Append an object to end. If the object is a string, appends a
        :class:`Word <Word>` object.
        """
        if isinstance(obj, (str, bytes)):
            super(WordList, self).append(Word(obj))
        else:
            super(WordList, self).append(obj)

    def extend(self, iterable):
        """Extend WordList by appending elements from ``iterable``. If an element
        is a string, appends a :class:`Word <Word>` object.
        """
        for e in iterable:
            self.append(e)

class Sentence(object):
    """A sentence within a AmharicDocument.

    :param sentence: A string, the raw sentence.
    :param start_index: An int, the index where this sentence begins
                        in a TextBlob. If not given, defaults to 0.
    :param end_index: An int, the index where this sentence ends in
                        a TextBlob. If not given, defaults to the
                        length of the sentence - 1.
    """

    def __init__(self, sentence, start_index=0, end_index=None, clean_sentence=None):
        self.sentence = sentence
        self.start = self.start_index = start_index
        self.end = self.end_index = end_index or len(sentence) - 1
        self.clean_sentence = clean_sentence
    
    def __repr__(self):
        '''Returns a string representation for debugging.'''
        class_name = self.__class__.__name__
        return f'{class_name}("{self.clean_sentence}")'
    
    @property
    def dict(self):
        '''The dict representation of this sentence.'''
        return {
            'raw_sentence': self.sentence,
            'clean_sentence': self.clean_sentence,
            'start_index': self.start_index,
            'end_index': self.end_index,   
        }
    


from collections import defaultdict
from functools import cached_property
from typing import Callable, List, Optional

from etltk.tokenize.am import EthiopicSentenceTokenizer, word_tokenize
from etltk.tokenize.space import whitespace_tokenize

from etltk.common.doc import (
    Sentence, 
    WordList
)

from .punctuation import ASSCII_PUNCT, ETHIOPIC_PUNCT, NO_ABBREV_ASSCII_ETHIOPIC_PUNCTS

from .preprocessing import (
    remove_links,
    remove_tags,
    remove_emojis,
    remove_email,
    remove_special_characters,
    remove_digits,
    remove_english_chars,
    remove_arabic_chars,
    remove_chinese_chars,
    remove_whitespaces,
    remove_ethiopic_digits,
    remove_ethiopic_punct,
    remove_non_ethiopic,
    remove_punct,
)
from .normalizer import (
   normalize_punct,
   normalize_shortened,
   normalize_char,
   normalize_labialized
)

DEFAULT_PIPELINE: List[Callable] = [
    remove_links,
    remove_tags,
    remove_emojis,
    remove_email,
    remove_special_characters,
    remove_digits,
    remove_ethiopic_digits,
    # TODO: text = remove_ethiopic_dates(text)
    remove_english_chars,
    remove_arabic_chars,
    remove_chinese_chars
]

def clean_amharic(text: str,  abbrev=False, pipeline: Optional[List[Callable]] = None):
    """ Returns a preprocessed copy of *text*,
    by executing a series of data preprocessing steps defined in pipeline. 

    Args:
        text (str): _description_
        abbrev (bool, optional): _description_. Defaults to False.
        pipeline (Optional[List[Callable]], optional): _description_. Defaults to None.

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """
    if text is None:
        raise ValueError("clean_amharic: `text` can't be `None`")
    
    if pipeline is None:
        pipeline = DEFAULT_PIPELINE
        
    for pipe_func in pipeline:
        text = pipe_func(text)
    
    text = normalize_punct(text)
    if not abbrev:
        text = normalize_shortened(text)
    
    text = remove_punct(text, abbrev=abbrev)
    
    text = remove_whitespaces(text)

    if isinstance(text, str):
        processed_text = text
    else:
        processed_text = " ".join(text)
    return processed_text

def normalize(text: str) -> str:
    """The function for all default amharic normalization. 
    Nomalizes an input text by executing a series of nomalization functions specified in the argument.
    
    Labialized Character Normalzation such as ሞልቱዋል to ሞልቷል
    Short Form Expansion such as ጠ/ሚ to ጠቅላይ ሚኒስተር.
    Punctuation Normalization such as :: to ።.
    Character Level Normalization such as ጸሀይ and ፀሐይ.
    """
    
    if text is None:
        raise ValueError("normalize: `text` can't be `None`")
   
    normalized_text = normalize_labialized(text)
    normalized_text = normalize_shortened(normalized_text)
    normalized_text = normalize_punct(normalized_text)
    return normalize_char(normalized_text)
            
class AmharicDocument(object):
    def __init__(self, text, tokenizer=None, clean_text=True):
        if not isinstance(text, (str, bytes)):
            raise TypeError('The `text` argument passed to `__init__(text)` '
                            f'must be a string, not {type(text)}')
            
        if not len(text.strip()):
            raise ValueError("AmharicDocument: `text` can't be `Empty String`")
        
        if clean_text:
            self.cleaned = clean_amharic(text)
        
        self.raw = self.string = text

    def __repr__(self):
        '''Returns a string representation for debugging.'''
        class_name = self.__class__.__name__
        text = self.cleaned
        return f'{class_name}("{text}")'
    
    @cached_property
    def tokens(self):
        """Return a list of tokens. This includes
        An individual token – i.e. a word, punctuation symbol, whitespace.

        :returns: A :class:`WordList <WordList>` of word tokens.
        """
        return WordList(word_tokenize(self.raw, return_expand=True, return_word=False))
    
    @cached_property
    def words(self):
        """Return a list of word tokens. This excludes punctuation characters.
        If you want to include punctuation characters, access the ``tokens``
        property.

        :returns: A :class:`WordList <WordList>` of word tokens.
        """
        return WordList(word_tokenize(self.raw, return_expand=True, return_word=True))
    
    @cached_property
    def sentences(self):
        """Return list of :class:`Sentence <Sentence>` objects."""
        return self._create_sentence_objects()
    
    @cached_property
    def word_counts(self):
        """Dictionary of word frequencies in this text.
        Count Tokenized Words
        """
        counts = defaultdict(int)
        stripped_words = [word for word in self.words]
        for word in stripped_words:
            counts[word] += 1
        return counts
    
    def ngrams(self, n=3):
        """Return a list of n-grams (tuples of n successive words) for this
        blob.

        :rtype: List of :class:`WordLists <WordList>`
        """
        if n <= 0:
            return []
        grams = [WordList(self.words[i:i + n])
                            for i in range(len(self.words) - n + 1)]
        return grams
    
    def _create_sentence_objects(self):

        sentence_objects = []        
        sentences = EthiopicSentenceTokenizer().tokenize(self.raw)
        
        # Since `EthiopicSentenceTokenizer` normalizes puctuation
        # `raw` has to be also normalized to extact spans
        punct_norm_text = normalize_punct(self.raw)
        
        char_index = 0  # Keeps track of character index within the blob
        for sent in sentences:
            # Compute the start and end indices of the sentence
            # within the blob
            start_index = punct_norm_text.index(sent, char_index)
            char_index += len(sent)
            end_index = start_index + len(sent)
            
            # Split into words by white space                
            expanded_words = normalize_shortened(sent)
            
            # Split into words by white space
            # Remove extra spaces, tabs, and new lines
            whitespaced_tokens = whitespace_tokenize(expanded_words)
            
            # remove non ethiopic chars and ethiopic punctuations
            ethiopic_tokens = [
                remove_non_ethiopic(token) for token in whitespaced_tokens if len(remove_non_ethiopic(token).strip())
            ]
            stripped_ethiopic_tokens = [
                remove_ethiopic_punct(token) for token in ethiopic_tokens if len(remove_ethiopic_punct(token).strip())
            ]        
            clean_sentence = " ".join(stripped_ethiopic_tokens)
            
            # Sentences share the same models as their parent blob
            s = Sentence(sent, start_index=start_index, end_index=end_index, clean_sentence=clean_sentence)
            sentence_objects.append(s)
        return sentence_objects     


# coding: utf8
from __future__ import unicode_literals

# Standard libraries
import re
import unicodedata
from typing import List

# etltk libraries
from .line import LineTokenizer
from .regexp import RegexpTokenizer
from .space import whitespace_tokenize
from ..lang.am.punctuation import ETHIOPIC_SENT_PUNCT, NO_ABBREV_ASSCII_ETHIOPIC_PUNCTS
from ..lang.am.preprocessing import remove_ethiopic_punct, remove_non_ethiopic
from ..lang.am.normalizer import normalize_punct, normalize_shortened

class EthiopicSentenceTokenizer(object):
    def __init__(self, sent_end_chars: List[str] = None):
        """This tokenizer segmented the sentence on the basis of the
        ethiopic sentence punctuation marks i.e (`፤`, `፥`, `።`).
        
        Uses an instance of LineTokenizer and RegexpTokenizer.
       
        Args:
            sent_end_chars (List[str], optional): list of sentences-ending punctuation marks. 
            Defaults to None.
        """            
        if sent_end_chars:
            self.sent_end_chars = sent_end_chars
            self.sent_end_chars_regex = "|".join(self.sent_end_chars)
        else:
            self.sent_end_chars_regex = ETHIOPIC_SENT_PUNCT # Default sent_end_chars ("፤", "፥", "።")
        
        self.pattern = rf"(?<=[{self.sent_end_chars_regex}])\s"
    
    def tokenize(self, text: str) -> List[str]:
        """Method for tokenizing sentences with regular expressions.
        Tokenize a text into a sequence of sentenece using end sentenece punctuation characters as a separator.
        
        Args:
            text (str): text to be tokenized into sentences

        Returns:
            List[str]: tokenized sentence list
        """
        lines = LineTokenizer().tokenize(text)
        
        # punctuation normalization 
        # :: -> ። 
        punct_norm_text = normalize_punct(text)
        
        sentences = [
            sent for sent in RegexpTokenizer(pattern=self.pattern, gaps=True).tokenize(punct_norm_text) if len(sent.strip())
        ]
        return sentences


# Sentence tokenizer.
def sent_tokenize(text: str) -> List[str]:
    """ Return a sentence-tokenized copy of *text*
    uses an instance of EthiopicSentenceTokenizer.

    Args:
        text (str): text to split into sentences

    Returns:
        List[str]: _description_
    """
    sentences = EthiopicSentenceTokenizer().tokenize(text)
    
    stripped_sentences: List = []
    for sent in sentences:
        # Split into words by white space                
        expanded_words = normalize_shortened(sent)
        
        # Split into words by white space
        # Remove extra spaces, tabs, and new lines
        whitespaced_tokens = whitespace_tokenize(expanded_words)
        
        # remove_non_ethiopic and ethiopic punctuations
        ethiopic_tokens = [
            remove_non_ethiopic(token) for token in whitespaced_tokens if len(remove_non_ethiopic(token).strip())
        ]
        stripped_ethiopic_tokens = [
            remove_ethiopic_punct(token) for token in ethiopic_tokens if len(remove_ethiopic_punct(token).strip())
        ]
        
        stripped_sentences.append(" ".join(stripped_ethiopic_tokens))
    
    return stripped_sentences

def wordpunct_tokenize(text: str) -> List[str]:
    # punctuation = "!\"#$%&'()*+,-:;<=>?@[\]^`{|}~።፤;፦፥፧፨፠፣"
    punctuation = "".join(NO_ABBREV_ASSCII_ETHIOPIC_PUNCTS)
    sentence_out = ""
    for character in text:
        if character in punctuation:
            sentence_out += " %s " % character
        else:
            sentence_out += character

    return sentence_out.split()

def word_tokenize(text: str, return_expand=True, return_word=True) -> List[str]:
    """_summary_

    Args:
        text (str): _description_
        return_expand (bool, optional): _description_. Defaults to True.
        return_word (bool, optional): _description_. Defaults to False.

    Returns:
        List[str]: _description_
    """
    word_tokens = wordpunct_tokenize(text)
    
    # Expands shortened characters
    if return_expand:
        expanded_words = normalize_shortened(" ".join(word_tokens))
        word_tokens = whitespace_tokenize(expanded_words)
    
    if return_word:
        # text cleaning, 
        # remove_non_ethiopic and ethiopic punctuations
        word_tokens = [
            remove_non_ethiopic(token) for token in word_tokens if len(remove_non_ethiopic(token).strip())
        ]
        word_tokens = [
            remove_ethiopic_punct(token) for token in word_tokens if len(remove_ethiopic_punct(token).strip())
        ]
        return word_tokens
    
    return word_tokens


# coding=utf-8
#
# Standard libraries
import re

# etltk libraries

class RegexpTokenizer():
    """A tokenizer that splits a string using a regular expression, which
    matches either the tokens or the separators between tokens.

    """
    def __init__(
        self,
        pattern,
        gaps=False,
        discard_empty=True,
        flags=re.UNICODE | re.MULTILINE | re.DOTALL,
    ):
        # If they gave us a regexp object, extract the pattern.
        pattern = getattr(pattern, "pattern", pattern)

        self._pattern = pattern
        self._gaps = gaps
        self._discard_empty = discard_empty
        self._flags = flags
        self._regexp = None


    def _check_regexp(self):
        if self._regexp is None:
            self._regexp = re.compile(self._pattern, self._flags)

    def tokenize(self, text):
        self._check_regexp()
        # If our regexp matches gaps, use re.split:
        if self._gaps:
            if self._discard_empty:
                return [tok for tok in self._regexp.split(text) if tok]
            else:
                return self._regexp.split(text)

        # If our regexp matches tokens, use re.findall:
        else:
            return self._regexp.findall(text)

    def __repr__(self):
        return "{}(pattern={!r}, gaps={!r}, discard_empty={!r}, flags={!r})".format(
            self.__class__.__name__,
            self._pattern,
            self._gaps,
            self._discard_empty,
            self._flags,
        )



# coding: utf8
from __future__ import unicode_literals

split_chars = lambda char: list(char.strip().split(" "))
merge_chars = lambda char: char.strip().replace(" ", "|")
group_chars = lambda char: char.strip().replace(" ", "")

_ethiopic = r"\u1200-\u137F"

_ethiopic_punct = (
    r"፠ ፡ ። ፣ ፤ ፥ ፦ ፧ ፨"
)

_ethiopic_sent_punct = (
    r"፤ ፥ ።"
)

_ethiopic_abbrev_punct = (
    r". /"
)

_asscii_punct = (
    r' ! " \# \$ % \& \' \( \) \* \+ , \- \. / : ; < = > \? @ \[ \\ \] \^ _ ` \{ \| \} \~ \''
)

ASSCII_PUNCT = merge_chars(_asscii_punct)
ETHIOPIC_PUNCT = merge_chars(_ethiopic_punct)
ETHIOPIC_ABBREV_PUNCT = merge_chars(_ethiopic_abbrev_punct)
ETHIOPIC_SENT_PUNCT = merge_chars(_ethiopic_sent_punct)

_asscii_ethiopic_puncts =  ASSCII_PUNCT + ETHIOPIC_PUNCT

table = str.maketrans('', '', _ethiopic_abbrev_punct.strip().replace(" ", ""))

# list of all punctuation marks except the ". /" (because we want
# to keep as one token the words like ዓ.ም, ጠ/ሚ etc.
NO_ABBREV_ASSCII_ETHIOPIC_PUNCTS =  "".join([punct.translate(table) for punct in _asscii_ethiopic_puncts])


# Whitespace Tokenizer
def whitespace_tokenize(text: str) -> str:
    """Tokenize a string on whitespace (space, tab, newline).
    """
    text = text.strip()
    if not text:
        return []
    # Remove extra spaces, tabs, and new lines
    tokens = text.split()
    return tokens
